@@ -33,6 +33,7 @@ from django.utils.safestring import mark_safe
 from django.utils.timesince import timeuntil
 from django.utils.translation import ungettext, ugettext, ugettext_lazy as _
 from django.contrib.sites.models import get_current_site
+from django.core.exceptions import ValidationError
 
 import chroniker.constants as const
 import chroniker.utils
@@ -118,6 +119,7 @@ class JobHeartbeatThread(threading.Thread):
         self.job_id = job_id
         self.lock = lock
         self.lock_file = tempfile.NamedTemporaryFile()
+        self.original_pid = os.getpid()
         set_current_job(job_id)
         set_current_heartbeat(self)
         threading.Thread.__init__(self, *args, **kwargs)
@@ -128,6 +130,14 @@ class JobHeartbeatThread(threading.Thread):
         """
         check_freq_secs = 5
         while not self.halt:
+            
+            # If the current PID doesn't match the one we started with
+            # then that means we were forked by a subprocess launched by the
+            # job. In this case, we're a clone, so immediately exit so we don't
+            # conflict with the thread in the original process.
+            if os.getpid() != self.original_pid:
+                return
+            
             #print 'Heartbeat check...'
             self.lock_file.seek(0)
             self.lock_file.write(str(time.time()))
@@ -166,8 +176,22 @@ class JobHeartbeatThread(threading.Thread):
         self.lock_file.close()
         
     def update_progress(self, total_parts, total_parts_complete, lock=True):
+        """
+        JobHeartbeatThread
+        """
         if lock:
             self.lock.acquire()
+#        print '!'*80
+#        import thread
+#        thread_ident = thread.get_ident()
+#        print 'self1:',self
+#        print 'id1:',id(self)
+#        print 'thread1:',thread_ident
+#        print 'job_id1:',self.job_id
+#        print 'pid1:',os.getpid()
+#        print 'total_parts1:',total_parts
+#        print 'total_parts_complete1:',total_parts_complete
+#        print '!'*80
         Job.objects.filter(id=self.job_id).update(
             total_parts = total_parts,
             total_parts_complete = total_parts_complete,
@@ -436,6 +460,13 @@ class Job(models.Model):
         blank=True,
         help_text=_("Space separated list; e.g: arg1 option1=True"))
     
+    raw_command = models.CharField(
+        _("raw command"),
+        max_length=1000,
+        blank=True,
+        null=True,
+        help_text=_("A raw shell command to run. This is mutually exclusive with `command`."))
+    
     enabled = models.BooleanField(
         default=True,
         help_text=_('''If checked, this job will be run automatically according
@@ -689,8 +720,8 @@ class Job(models.Model):
         td = timezone.now() - self.last_run_start_timestamp
         if not progress_ratio:
             return
-        total_sec = td.seconds/progress_ratio
-        remaining_sec = total_sec - td.seconds
+        total_sec = td.total_seconds()/progress_ratio
+        remaining_sec = total_sec - td.total_seconds()
         return remaining_sec
     
     @property
@@ -709,7 +740,36 @@ class Job(models.Model):
     estimated_completion_datetime_str.short_description = 'ETC'
     estimated_completion_datetime_str.help_text = 'Estimated time of completion'
     
+    def clean(self, *args, **kwargs):
+        cmd1 = (self.command or '').strip()
+        cmd2 = (self.raw_command or '').strip()
+        if cmd1 and cmd2:
+            raise ValidationError({
+                'command':[
+                    'Either specify command or raw command, but not both.',
+                ],
+                'raw_command':[
+                    'Either specify command or raw command, but not both.',
+                ]
+            })
+        elif not cmd1 and not cmd2:
+            raise ValidationError({
+                'command':[
+                    'Either command or raw command must be specified.',
+                ],
+                'raw_command':[
+                    'Either command or raw command must be specified.',
+                ]
+            })
+        super(Job, self).clean(*args, **kwargs)
+    
+    def full_clean(self, exclude=None, validate_unique=None, *args, **kwargs):
+        return self.clean(*args, **kwargs)
+    
     def save(self, *args, **kwargs):
+        
+        self.full_clean()
+        
         if self.enabled:
             if self.pk:
                 j = Job.objects.get(pk=self.pk)
@@ -911,7 +971,7 @@ class Job(models.Model):
         """
         return self.is_due() and self.dependencies_met(running_ids=running_ids)
     
-    def run(self, check_running=True, *args, **kwargs):
+    def run(self, check_running=True, force_run=False, *args, **kwargs):
         """
         Runs this ``Job``.  A ``Log`` will be created if there is any output
         from either stdout or stderr.
@@ -923,7 +983,11 @@ class Job(models.Model):
         dueness and don't want our current run status to give an incorrect
         reading.
         """
-        if self.enabled:
+        print 'force_run:',force_run
+        if force_run:
+            self.handle_run(*args, **kwargs)
+            return True
+        elif self.enabled:
             if not self.dependencies_met():
                 # Note, this will cause the job to be re-checked
                 # the next time cron runs.
@@ -953,19 +1017,21 @@ class Job(models.Model):
         run_start_datetime = timezone.now()
         last_run_successful = False
         
-        stdout = sys.stdout
-        stderr = sys.stderr
-        if self.log_stdout:
-            stdout = chroniker.utils.TeeFile(
-                sys.stdout,
-                auto_flush=True,
-                queue=stdout_queue)
-        if self.log_stderr:
-            stderr = chroniker.utils.TeeFile(
-                sys.stderr,
-                auto_flush=True,
-                queue=stderr_queue)
+        original_pid = os.getpid()
+        
         try:
+            stdout = sys.stdout
+            stderr = sys.stderr
+            if self.log_stdout:
+                stdout = chroniker.utils.TeeFile(
+                    sys.stdout,
+                    auto_flush=True,
+                    queue=stdout_queue)
+            if self.log_stderr:
+                stderr = chroniker.utils.TeeFile(
+                    sys.stderr,
+                    auto_flush=True,
+                    queue=stderr_queue)
     
             # Redirect output so that we can log it if there is any
             ostdout = sys.stdout
@@ -1018,9 +1084,20 @@ class Job(models.Model):
             success = True
             try:
                 logger.debug("Calling command '%s'" % self.command)
-                call_command(self.command, *args, **options)
+                if self.raw_command:
+                    retcode = subprocess.call(
+                        self.raw_command,
+                        shell=True,
+                        stdout=sys.stdout,
+                        stderr=sys.stderr)
+                else:
+                    call_command(self.command, *args, **options)
                 logger.debug("Command '%s' completed" % self.command)
+                if original_pid != os.getpid():
+                    return
             except Exception, e:
+                if original_pid != os.getpid():
+                    return
                 # The command failed to run; log the exception
                 t = loader.get_template('chroniker/error_message.txt')
                 c = Context({
@@ -1091,6 +1168,11 @@ class Job(models.Model):
                             
         finally:
             
+            if original_pid != os.getpid():
+                # We're a clone of the parent job, so exit immediately
+                # so we don't conflict.
+                return
+            
             # Redirect output back to default
             sys.stdout = ostdout
             sys.stderr = ostderr
@@ -1115,7 +1197,7 @@ class Job(models.Model):
                     stderr_str = unicode(stderr_str, 'utf-8', 'replace')
             
             run_end_datetime = timezone.now()
-            duration_seconds = (run_end_datetime - run_start_datetime).seconds
+            duration_seconds = (run_end_datetime - run_start_datetime).total_seconds()
             log = Log.objects.create(
                 job = self,
                 run_start_datetime = run_start_datetime,
@@ -1251,7 +1333,7 @@ class Log(models.Model):
         if self.run_start_datetime and self.run_end_datetime:
             assert self.run_start_datetime <= self.run_end_datetime, 'Job must start before it ends.'
             time_diff = (self.run_end_datetime - self.run_start_datetime)
-            self.duration_seconds = time_diff.seconds
+            self.duration_seconds = time_diff.total_seconds()
         super(Log, self).save(*args, **kwargs)
     
     def duration_str(self):
